@@ -1,19 +1,21 @@
 from pathlib import Path
 import typing as T
 from uuid import uuid4
-
+import logging
 
 from .datatypes import HANDLE
 from .bass_module import Bass
 from .bass_channel import BassChannel
 from .bass_stream import BassStream
 
+log = logging.getLogger(__name__)
 
 class Song:
 
     _handle: HANDLE
     _handle_length: float # Seconds
     _handle_position: float # Seconds
+    file_path: Path
 
     def __init__(self, file_path: T.Union[str, Path]):
         super(Song, self).__init__()
@@ -21,7 +23,12 @@ class Song:
 
         self._id = uuid4().hex
         self.file_path = Path(file_path)
-        assert self.file_path.exists() and self.file_path.is_file(), f"{self.file_path} doesn't exist or is not a file"
+        if self.file_path.exists() is False:
+            raise ValueError(f"{file_path} doesn't exist")
+
+        if self.file_path.is_file() is False:
+            raise ValueError(f"{file_path=} is not a valid file")
+
 
         self._handle = None
         self._handle_length = None # Length in seconds
